#include <memory>
#include <stdexcept>
#include "kernels.hpp"
#include "strategy.hpp"
#include "utils.hpp"

Kernel::Kernel(const std::string& name, Kernel::StrategyFunction strategy_function, Kernel::PreparationFunction preparation_function)
  : name_(name), strategy_function_(std::move(strategy_function)), preparation_function_(std::move(preparation_function)) {}

void Kernel::prepare() const {
  preparation_function_();
}

void Kernel::execute(int num_threads_or_tasks, int kernel_tripcount) const {
  strategy_function_(0, kernel_tripcount, num_threads_or_tasks);
}

void KernelRegistry::register_kernel(const std::string& name, KernelBuilder factory) {
  registry_.emplace(name, std::move(factory));
}

Kernel KernelRegistry::load_kernel(const std::string& name) const {
  auto it = registry_.find(name);
  if (it == registry_.end()) {
    throw std::invalid_argument("Kernel not found: " + name);
  }
  return it->second();
}

std::vector<std::string> KernelRegistry::list_available_kernels() const {
  std::vector<std::string> kernel_names;
  for (const auto& entry : registry_) {
    kernel_names.push_back(entry.first);
  }
  return kernel_names;
}

// New kernels go here, each can have it's own set of arguments and initializations
// execute() contains the full kernel code minus an outer for loop (i=start, i<end, ++i), 
// defined in the respective parallelization strategy
void initialize_registry(KernelRegistry* registry, std::string strategy_name) {

  // STREAM TRIAD
  registry->register_kernel("stream_triad", [&]() {
    auto a = std::make_shared<std::vector<float>>();
    auto b = std::make_shared<std::vector<float>>();
    auto c = std::make_shared<std::vector<float>>();
    
    auto prepare = [=]() {
      a->resize(VECTOR_SIZE);
      b->resize(VECTOR_SIZE);
      c->resize(VECTOR_SIZE);
      initialize_vector(*b);
      initialize_vector(*c);
    };

    auto execute = [=](int kernel_start_idx, int kernel_end_idx, int num_threads_or_tasks) {
      strategy::execute_strategy(strategy_name, kernel_start_idx, kernel_end_idx, num_threads_or_tasks, [&](int i) {
        (*a)[i] = (*b)[i] + 0.5f * (*c)[i];
      });
    };

    return Kernel("stream_triad", execute, prepare);
  });

  // DAXPY
  registry->register_kernel("daxpy", [&]() {
    auto a = std::make_shared<std::vector<float>>();
    auto b = std::make_shared<std::vector<float>>();

    auto prepare = [=]() {
      a->resize(VECTOR_SIZE);
      b->resize(VECTOR_SIZE);
      initialize_vector(*b);
    };

    auto execute = [=](int kernel_start_idx, int kernel_end_idx, int num_threads_or_tasks) {
      strategy::execute_strategy(strategy_name, kernel_start_idx, kernel_end_idx, num_threads_or_tasks, [&](int i) {
        (*a)[i] += 0.5f * (*b)[i];
      });
    };

    return Kernel("daxpy", execute, prepare);
  });

}

