#include <iostream>
#include <chrono>
#include "kernels.hpp"

int main(int argc, char** argv) {

    if (argc != 4) {
        std::cerr << "Usage: " << argv[0] << " <kernel_name> <strategy> <num_threads_or_tasks>\n";
        return 1;
    }

    std::string kernel_name = argv[1];
    std::string strategy_name = argv[2];
    int num_threads_or_tasks = std::stoul(argv[3]);
    
    // registry contains a map of kernels generated from kernel builders for the selected parallelization strategy
    KernelRegistry registry;
    initialize_registry(&registry, strategy_name);
    
    try{ 
      // find kernel in unordered_map by it's name. prepare() allocates and initializes data structures needed for the selected kernel
      Kernel kernel = registry.load_kernel(kernel_name);
      kernel.prepare();

      // Time the kernel execution
      auto start_time = std::chrono::high_resolution_clock::now();
      
      // VECTOR_SIZE is a preprocessor variable to mimic the setup of STREAM
      kernel.execute(num_threads_or_tasks, VECTOR_SIZE);
      
      auto end_time = std::chrono::high_resolution_clock::now();
      std::chrono::duration<double, std::milli> duration = end_time - start_time;

      std::cout << "Kernel: " << kernel_name << "\n";
      std::cout << "Parallelization strategy: " << strategy_name << "\n";
      std::cout << "Number of threads / tasks: " << num_threads_or_tasks << "\n";
      std::cout << "Kernel execution time [ms]: " << duration.count() << "\n";
    } catch (const std::invalid_argument& e) {
        // If kernel name is invalid, list available kernels
        std::cerr << e.what() << "\n";
        std::cerr << "Available kernels are:\n";
        
        // List available kernels from registry
        for (const auto& kernel_name : registry.list_available_kernels()) {
            std::cerr << "  - " << kernel_name << "\n";
        }

        return 1;
    }
    return 0;
}
